(* Testing the FFTW interface *)

open Bigarray
open Printf

let complex a b = { Complex.re = a; Complex.im = b }


let n = 16
let fft = Fftw.create Fftw.forward n
and ifft = Fftw.create Fftw.backward ~normalize:true n

let () =
  let a = Array1.create complex64 c_layout n in
  let () =
    for i = 0 to n - 1 do
      a.{i} <- complex 1. 0.
    done in
  let fa = fft a in
  let ifa = ifft fa in
    for i = 0 to n - 1 do
      printf "(%g, %g) " fa.{i}.Complex.re fa.{i}.Complex.im
    done;
    printf "...\n";
    for i = 0 to n - 1 do
      printf "(%g, %g) " ifa.{i}.Complex.re ifa.{i}.Complex.im
    done;
    printf "...\n"

let () =
  let l = Fftw.wisdom_export_fold (fun c (i,l) -> (i+1, l @ [c])) (0,[]) in
    print_string (Fftw.wisdom_to_string()); print_newline();
    Gc.full_major();
    List.iter (fun c -> print_char c) (snd l); print_newline();
    Fftw.wisdom_from_string "(FFTW-2.1.3 (23 4 garbage))"


(* Multidimensional complex transforms *)

let n = 4
and layout = c_layout
let fft = Fftw.nd_create Fftw.forward layout [| n; n|]
and ifft = Fftw.nd_create Fftw.backward ~normalize:true layout [| n; n|]

let slice m no = Genarray.slice_left m [|no|]
(* let slice m no = Genarray.slice_right m [| no + 1 |] *)

let indices2, indices3, dims3 =
  if layout = (Obj.magic c_layout : 'a Bigarray.layout) then
    (fun i1 i2 -> [| i1; i2 |]) ,
    (fun no i1 i2 -> [| no; i1; i2 |]),
    (fun no i1 i2 -> [| no; i1; i2 |])
  else
    (fun i1 i2 -> [| i1 + 1; i2 + 1 |]) ,
    (fun no i1 i2 -> [| i1 + 1; i2 + 1; no + 1 |]),
    (fun no i1 i2 -> [| i1; i2; no |])

let print_matrix_2d m =
  printf "[|";
  for i1 = 0 to Genarray.nth_dim m 0 - 1 do
    for i2 = 0 to Genarray.nth_dim m 1 - 1 do
      let z = Genarray.get m (indices2 i1 i2) in
	printf " (%g,%g)" z.Complex.re z.Complex.im
    done;
    printf "\n  "
  done;
  printf " |]\n"

let () =
  let m = Genarray.create complex64 layout (dims3 2 n n) in
  let () = 
    for no = 0 to 1 do
      for i1 = 0 to n - 1 do
	for i2 = 0 to n - 1 do
	  Genarray.set m (indices3 no i1 i2) (complex (float (no + 1)) 0.)
	done
      done
    done in
  let fm = fft ~index_dims:1 m in
  let ffm = ifft ~index_dims:1 fm in
    for no = 0 to 1 do
      print_matrix_2d (slice m no)
    done;
    for no = 0 to 1 do
      print_matrix_2d (slice fm no)
    done;
    for no = 0 to 1 do
      print_matrix_2d (slice ffm no);
    done


(* Multidimensional real transforms *)

let rfft = Fftw.rnd_create Fftw.forward layout [| n; n|]
and irfft = Fftw.rnd_create Fftw.backward ~normalize:true layout [| n; n|]

let print_matrix_2d_float m =
  printf "[|";
  for i1 = 0 to Genarray.nth_dim m 0 - 1 do
    for i2 = 0 to Genarray.nth_dim m 1 - 1 do
	printf " %g" (Genarray.get m (indices2 i1 i2))
    done;
    printf "\n  "
  done;
  printf " |]\n"


let () =
  let m = Genarray.create float64 layout (dims3 2 n n) in
  let () =
    for no = 0 to 1 do
      for i1 = 0 to n - 1 do
	for i2 = 0 to n - 1 do
	  Genarray.set m (indices3 no i1 i2) (float (no + 1))
	done
      done
    done in
  let fm = rfft ~index_dims:1 m in
  let ffm = irfft ~index_dims:1 fm in
    for no = 0 to 1 do
      print_matrix_2d_float (slice m no)
    done;
    for no = 0 to 1 do
      print_matrix_2d (slice fm no)
    done;
    for no = 0 to 1 do
      print_matrix_2d_float (slice ffm no);
    done
